"""
@author: Adam Micolich
Updated by Jakob Seidl
This module does the input handling for the USB-6216, which is effectively a pair of analog outputs
and a set of 8 analog inputs. The output handling is done by a separate .py. 
"""

import pyneMeas.Instruments.Instrument as Instrument
import nidaqmx as nmx


@Instrument.enableOptions
class USB6216In(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    defaultInput = "inputLevel"
    defaultOutput = "None"

    def __init__(self, address,usbPort = 'Dev1'):
        super(USB6216In, self).__init__()
        self.dev = address
        self.usbPort = usbPort
        self.type ="USB6216"  #We can check each instrument for its type and react accordingly
        self.name = "myUSB6216"

        # Define the self.port which is the handle to the device
        if self.dev in [*range(8)]:
            self.port = f"{self.usbPort}/ai{self.dev}"
            
        else: raise ValueError(f'Please insert a valid Input port for the NIDaQ ranging from 0 to 7. You entered {self.dev}')
        
    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
         
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
        
    @Instrument.addOptionGetter("inputLevel")
    def _getInputLevel(self):
        with nmx.Task() as task:
            task.ai_channels.add_ai_voltage_chan(self.port)
            tempData = task.read()
        measInput = float(tempData)/self.scaleFactor        
        return measInput
    
    @Instrument.addOptionGetter("scaleFactor")
    def _getScaleFactor(self):
        return self.scaleFactor
    
    @Instrument.addOptionSetter("scaleFactor")
    def _setScaleFactor(self,scaleFactor):
        self.scaleFactor = scaleFactor
            
    def goTo(self,target,stepsize=0.001,delay=0.02):
        pass
            
    def close(self):
        pass


from pyneMeas.utility.GlobalMeasID import readCurrentSetup
from pyneMeas.utility.GlobalMeasID import readCurrentID
from pyneMeas.utility.GlobalMeasID import init
from pyneMeas.utility.GlobalMeasID import listIDs
from pyneMeas.utility.GlobalMeasID import setCurrentSetup


from pyneMeas.utility.SweepFunction import sweepAndSave as sweep

from pyneMeas.utility.utils import targetArray
from pyneMeas.utility.utils import runTest


"""
VERSION 4.0
@author: Donny
jakob.seidl@nanoelectronics.physics.unsw.edu.au
"""
from collections import Iterable

# Checks if val is iterable, but not a string
def isIterable(val):
    return isinstance(val, Iterable) and not isinstance(val, str)

def flatten(iterable):
    flattenedList = []
    for e1 in iterable:
        if isIterable(e1):
            for e2 in e1:
                flattenedList.append(e2)
        else:
            flattenedList.append(e1)
    return flattenedList

# Converts one element lists to just the element itself (e.g., ["a"] -> "a"),
# while leaving everything else untouched. This way we can have an "intuitive"
# format for options that only take a single argument, but still preserve
# support for multiple argument options
def unwrap(args):
    if isIterable(args) and len(args) == 1:
        return args[0]
    return args

class Instrument(object):
    # This contains all the available options an instrument has. This is
    # populated automatically for each individual instrument class with the
    # @addOptionSetter/Getter and @enableOptions decorators defined later in the file.
    # The key is the option name, while the values are: {
    #     "order": The relative order the option should be set in, when setting
    #              multiple options at once (only for _optionSetters)
    #     "fn": The actual class member function that sets the option
    # }
    _optionSetters = {}
    _optionGetters = {}

    def __init__(self):
        self._options = {}

    def get(self, option, forceCached = False):
        if not forceCached and option in type(self)._optionGetters:
            return unwrap(self._optionGetters[option]["fn"](self))
        elif option in type(self)._optionSetters:
            return unwrap(self._options[option])
        else:
            raise ValueError(
                "{} is not a valid option to get. {} are available".format(
                    option,
                    set(type(self)._optionSetters.keys() + type(self)._optionGetters.keys())
                )
            )

    def set(self, option, *args):
        if not option in type(self)._optionSetters:
            raise ValueError("{} is not a valid option. {} are available".format(option, type(self)._optionSetters.keys()))

        type(self)._optionSetters[option]["fn"](self, *args)
        self._options[option] = args

    def getOptions(self):
        options = self._options.copy()
        for option, args in options.items():
            options[option] = unwrap(args)
        for option in type(self)._optionGetters:
            options[option] = self.get(option)
        return options

    def setOptions(self, options):
        # First check all options exist
        options = list(options.items())
        for option, _ in options:
            if not option in type(self)._optionSetters:
                raise ValueError("{} is not a valid option. {} are available".format(option, type(self)._optionSetters.keys()))

        # Sort the options so they're set in the correct order
        options = sorted(options, key = lambda option: type(self)._optionSetters[option[0]]["order"])

        # Actually set the options
        for option, args in options:
            if not isIterable(args):
                args = (args,)
            self.set(option, *args)

    def read(self):
        return self.get(self.defaultInput)
# Having a "totalOptions" variable lets us assign a total order to the options,
# by giving each option an unique number. This way we can ensure an order to
# setting the options when setting multiple options at once, since some options
# might require being set before others
totalOptions = 0
def addOptionSetter(name):
    def decorator(optionSetter):
        global totalOptions
        totalOptions += 1
        optionSetter.optionInfo = {"name": name, "order": totalOptions}
        return optionSetter
    return decorator

def addOptionGetter(name):
    def decorator(optionGetter):
        optionGetter.optionInfo = {"name": name, "isGetter": True}
        return optionGetter
    return decorator

# Since we can't access the class until it's finished being created, we have to
# have a separate decorator over the entire class, so we can then set the
# _optionsSetters property with the available options
def enableOptions(instrument):
    instrument._optionSetters = {}
    instrument._optionGetters = {}

    for option in instrument.__dict__.values():
        if not hasattr(option, "optionInfo"):
            continue
        info = option.optionInfo
        if "isGetter" in info:
            instrument._optionGetters[info["name"]] = {"fn": option}
        else:
            instrument._optionSetters[info["name"]] = {"order": info["order"], "fn": option}

    return instrument

def closeInstruments(instrumentList1,instrumentList2=None):
    if instrumentList2 != None:
        List = flatten([instrumentList1,instrumentList2])
    else: List = instrumentList1

    for instrument in List:
        try:
            instrument.close()
        except:
            continue


# Defining a new instrument
# ------------------------
# import Instrument
#
# @Instrument.enableOptions # Actually populate _optionSetters
# class Keithley2401(Instrument.Instrument):
#    def __init__(self, address):
#        super(Keithley2401, self).__init__() # Call Instrument's __init__
#        # Other initialisation stuff, like connecting to the instrument
#
#    @Instrument.addOptionSetter("beepEnable")
#    def _setBeepEnable(self, enable):
#        # Actually enable or disable beep here
#
#    @Instrument.addOptionSetter("option2")
#    def _setOption2(self, value):
#        # etc
#
# # By defining "beepEnable" before "option2", "beepEnable" is guaranteed to be
# # set before "option2", when both options are passed into setOptions().
# # Basically, the order the @addOption appear in define the order they are set
# # when they appear in setOptions()

# Saving/loading options
# ----------------------
# How options are saved/loaded is up to the user (though we could add a default
# way of doing so to sweeper.defaultWriter)
#
# For example, we could save two instruments' options as JSON:
# import json
# options = open("options.json", "w")
# options.write(json.dumps(
#     {"keithley": keithley.getOptions(), "lockin": lockin.getOptions()},
#     indent = 4, sort_keys = True
# ))
#
# To load the values back:
# import json
# options = open("options.json", "r")
# options = json.loads(options.read())
# keithley.setOptions(options["keithley"])
# lockin.setOptions(options["lockin"])



# -*- coding: utf-8 -*-
"""
VERSION 4.0
@author: Jakob Seidl
jakob.seidl@nanoelectronics.physics.unsw.edu.au
"""

import tkinter as tk
from tkinter.filedialog import asksaveasfilename, askdirectory

import json
import numpy as np
import matplotlib


def array(start,target,stepsize,upDown = False):
    
    sign = 1 if (target>start) else -1
    sweepArray = np.arange(start,target+sign*stepsize,sign*stepsize)
    if upDown:
        sweepArrayBack = np.arange(target,start+-sign*stepsize,-sign*stepsize)
        sweepArray = np.concatenate((sweepArray,sweepArrayBack))
    return sweepArray

def targetArray(targetList,stepsize):
    arrayList = []
    stepsize = float(stepsize)
    for index, item in enumerate(targetList):

        if index == 0:
             pass
        elif index == 1:
            arrayList.append(array(targetList[index - 1], targetList[index], stepsize))
        elif (index > 1):
            arrayList.append(array(targetList[index-1],targetList[index],stepsize)[1:]) #in order to avoid doublicating numbers
    return np.concatenate((arrayList))

def fileDialog(initAd = "/"):
    root = tk.Tk()
    #root.lift()
    root.attributes('-topmost','true')   
    root.withdraw()
    fullPath = asksaveasfilename(initialdir = initAd)
    a = ''.join(fullPath).rfind("/")
    basePath = fullPath[0:a+1]
    fileName = fullPath[a+1:]
    return [basePath,fileName]

def runTest():
    import pyneMeas.Instruments as I
    import pyneMeas.utility as U
    T = I.TimeMeas()
    Noi = I.LinearNoiseGenerator()
    NoiSin = I.SineNoiseGenerator()

    Dct = {}
    Dct['basePath'] = "TempDat/"
    Dct['fileName'] = 'fileName'

    Dct['setters'] = {T: 'dummy'}

    Dct['readers'] = {T: 'time',
                      Noi: 'lin_noise',
                      NoiSin: 'sine_noise',
                      }

    Dct['sweepArray'] = range(100)
    df = U.sweep(
        Dct,
        delay=0.0,
        plotVars=[('time', 'lin_noise'),
                  ('time', 'sine_noise')
                  ],
        plotParams=[('o-', 'linear-linear')
            , ('go-', 'linear-linear')
                    ],
        saveCounter=100,
    )


def sendEmail(targetAddress, measurementName):
    """
    Sends and email from our group's shared account. Users that are not in our group will find these credentials empty. They need to be populated before the email can be sent.

    """
    import smtplib
    sent_from = 'enterEmail'
    password = 'enter Password'
    subject = 'Measurement finished'
    body = 'Eureka, your measurement >>>' + measurementName + '<<< has just finished!'
    email_text = """\  
    From: %s  
    To: %s  
    Subject: %s

    %s
    """ % (sent_from, targetAddress, subject, body)
    mail = smtplib.SMTP('smtp.gmail.com:587')  # or 587
    mail.ehlo()
    mail.starttls()
    mail.login(sent_from, password)
    mail.sendmail(sent_from, targetAddress, email_text)
    print('Email sent to: ' + targetAddress)
    mail.close()


# -*- coding: utf-8 -*-
"""
@author: Jakob Seidl, Nanoelectronics Group UNSW Sydney
"""
#Electrometer



import pyvisa as visa
import pyneMeas.Instruments.Instrument as Instrument
import time
import math
@Instrument.enableOptions
class Keithley6517A(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    #defaultOutput = "sourceLevel"
    defaultInput = "senseLevel"
    
    def __init__(self, address):
        super(Keithley6517A, self).__init__()
        self.dev = visa.ResourceManager().open_resource("GPIB0::"+str(address)+"::INSTR")
        print(self.dev.query("*IDN?")) # Probably should query and check we have the right device
        self.type ="Keithley6517A" 
        self.name = 'myElectrometer'
        self.senseMode = 'current'
        self.zeroCheckEnable = True
        
        self.autoRange = False       
        self.possibleCurrRanges = [20E-12,200E-12,2E-9,20E-9,200E-9,2E-6,20E-6,200E-6,2E-3,20E-3]
        #self.possibleCurrAutoRanges = [200E-12,200E-9,2E-6,200E-6,2E-3,20E-3] #default for all ranges
        self.possibleCurrAutoRanges = [2E-9,200E-9,2E-6,20E-6,2E-3,20E-3] #A subset of 'self.possibleCurrRanges'!!
        self.possibleVoltRanges = [2,20,200]
        self.rangeIndex =  self.possibleCurrRanges.index(self._getSenseRange())          
        self.rangeDic = {  #we shift the lower intervals 30% (-> 0.7*number) downwards so they dont overlap with the bounds when we shift up.
            0:[15E-18,1.8E-9],
            1:[1*1E-9,180*1E-9],
            2:[130E-9,1.8*1E-6],
            3:[0.7*1E-6,17E-6],
            4:[15*1E-6,1.7E-3],
            5:[1.5*1E-3,20E-3],
            }
    
    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
    
    @Instrument.addOptionSetter("zeroCheck")    
    def _setZeroCheckEnable(self,enable):
        self.dev.write("SYST:ZCH "+("1" if enable else "0"))
        self.zeroCheckEnable = enable
        time.sleep(1)
        
    @Instrument.addOptionGetter("zeroCheck")    
    def _getZeroCheckEnable(self):
        zeroCheck = self.dev.query("SYST:ZCH?")
        if zeroCheck == u'1\n': 
            return True
        elif zeroCheck == u'0\n': 
            return False
        else:
            raise RuntimeError("Unknown zeroCheck mode.".format(zeroCheck))
    
    @Instrument.addOptionSetter("senseMode")
    def _setSenseMode(self,mode):     
        if mode == "voltage": 
            self.dev.write("SENS:FUNC 'VOLT'")
            self.senseMode = "voltage"
        elif mode == "current": 
            self.dev.write("SENS:FUNC 'CURR'")
            self.senseMode = "current"
        else:
            raise RuntimeError("\"{}\" is not a valid current/voltage measurement mode. Choose between 'voltage' or 'current'".format(mode))
    
    @Instrument.addOptionGetter("senseMode")
    def _getSenseMode(self):
        mode = self.dev.query("CONF?")
        if mode == u'"VOLT:DC"\n': 
            self.senseMode = 'voltage'
            return "voltage"
        elif mode == u'"CURR:DC"\n': 
            self.senseMode = 'current'
            return "current"
        else:
            raise RuntimeError("Unknown senseMode".format(mode))  


        
    @Instrument.addOptionGetter("senseLevel")
    def _getSenseLevel(self):
        if self.zeroCheckEnable:
            self.set('zeroCheck',False)
        
        #tempData = self.dev.query("READ?")
        tempData = self.dev.query(':FETCh?')
        divideIndex = tempData.find(",")
        result = float(tempData[0:divideIndex-4])
        if abs(result)> 1E10:
            result = float('nan')
        if self.autoRange:
            self._autoSensitivity(result)
        
        return result
    
    
    @Instrument.addOptionSetter("autoRange")
    def _setAutoRange(self,enable):
        if enable in (True,False):
            self.autoRange = enable
            if self._getSenseRange() in self.possibleCurrAutoRanges:
                self.rangeIndex =  self.possibleCurrAutoRanges.index(self._getSenseRange())
                self._setSenseRange(self.possibleCurrAutoRanges[self.rangeIndex]) #Not sure
                time.sleep(0.5)#Not sure
            else:
                self._setSenseRange(self.possibleCurrAutoRanges[-1])
                self.rangeIndex = self.possibleCurrAutoRanges.index(self._getSenseRange())
                self._setSenseRange(self.possibleCurrAutoRanges[self.rangeIndex]) #Not sure
                time.sleep(0.5)#Not sure
        else:
            raise RuntimeError("Unknown autoRange mode".format(enable))  
    
    @Instrument.addOptionGetter("autoRange")
    def _getAutoRange(self):
        return self.autoRange
    
    @Instrument.addOptionGetter("senseRange")
    def _getSenseRange(self):
        if self.senseMode == 'voltage':
            return float(self.dev.query(":VOLT:RANG?"))
        elif self.senseMode =='current':
            return float(self.dev.query(":CURR:RANG?"))  
        
    @Instrument.addOptionSetter("senseRange")
    def _setSenseRange(self,senseRange):

        if self.senseMode == 'current':
            if senseRange in self.possibleCurrRanges:
                self.dev.write(":CURR:RANG "+str(senseRange))
            else:
                raise RuntimeError("\"{}\" is not a valid current measurement range for the Keithley6517A electrometer".format(senseRange)+
                 "Possible values are \"{}\"".format(self.possibleCurrRanges))
        
        elif (self.senseMode =='voltage' and senseRange in self.possibleVoltRanges):
             self.dev.write(":VOLT:RANG "+str(senseRange))
             
    def _autoSensitivity(self,givenValue):
            R = abs(givenValue)
#            print('#############\n')
#            print('current value is'+str(R)+'\n')
#            print('current range index is: '+str(self.rangeIndex)+'\n')
#            print('the boundaries here are: '+str(self.rangeDic[self.rangeIndex][0]) +'and' + str(self.rangeDic[self.rangeIndex][1])+'\n')
#            print('#############\n')
            if (self.rangeDic[self.rangeIndex][0] >= R): #If current is below the current lower bound: Move one range down!  #and not math.isnan(R)
                a = self.rangeIndex -1
                if a in range(len(self.rangeDic)):
                    self.rangeIndex = a
                    print(a)
                    self._setSenseRange(self.possibleCurrAutoRanges[self.rangeIndex])
                    time.sleep(0.5)
            elif (R > self.rangeDic[self.rangeIndex][1]): #If current is above the current higher bound: Move one range up!  
                a = self.rangeIndex +1
                if a in range(len(self.rangeDic)):
                    self.rangeIndex = a
                    print(a)
                    self._setSenseRange(self.possibleCurrAutoRanges[self.rangeIndex])
                    time.sleep(0.5)
            else:
                if math.isnan(R): #when we measure a nan, better increase the range! This de=escalates the autorange when values change too quickly for the electrometer                   
                    a = self.rangeIndex +1
                    if a in range(len(self.rangeDic)):
                        self.rangeIndex = a
                        print(a)
                        self._setSenseRange(self.possibleCurrAutoRanges[self.rangeIndex])
                        time.sleep(0.5)    

    @Instrument.addOptionSetter("NPLC")
    def _setNPLC(self,NPLCs):
        if (NPLCs >= 0.01 and NPLCs <= 10):
            if self._getSenseMode() == "current":
                self.dev.write('CURR:NPLC '+str(NPLCs))
            elif self._getSenseMode() == 'voltage':
                self.dev.write('VOLT:NPLC '+str(NPLCs))
        else:
            raise RuntimeError("\"{}\" is not a valid NPLC input for the Keithley6517A electrometer".format(NPLCs)+
                 "Choose a value between 0.01 and 10.")

    @Instrument.addOptionGetter("NPLC")
    def _getNPLC(self):
        if self._getSenseMode() == 'current':
            NPLCs = self.dev.query('CURR:NPLC?')
            return float(NPLCs)
        else:
            NPLCs = self.dev.query('VOLT:NPLC?')
            return float(NPLCs)
    def close(self):
        self.set('zeroCheck',True)
        self.dev.close()
        

#[:SENSe[1]]:VOLTage[:DC]:RANG:AUTO
# [:SENSe[1]]:FUNC 'VOLT'  #Maybe this is the real measure mode not the one defined above...


# -*- coding: utf-8 -*-
"""

@author: Jakob Seidl@Nanoelectronics Group UNSW Sydney
"""
import pyvisa as visa
import pyneMeas.Instruments.Instrument as Instrument
import numpy as np
import time
@Instrument.enableOptions
class Keithley2401(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    defaultOutput = "sourceLevel"
    defaultInput = "senseLevel"

    def __init__(self, address):
        super(Keithley2401, self).__init__()
        self.dev = visa.ResourceManager().open_resource("GPIB0::"+str(address)+"::INSTR")
        print(self.dev.query("*IDN?")) # Probably should query and check we have the right device
        
        #self.dev.write("*RST")
        # self.dev.write("*CLS")
        self.type ="Keithley2401"  #We can check each instrument for its type and react accordingly
        self.name ='myKeithley2401'
        self.scaleFactor = 1.0
        self.currentSourceSetpoint = float('nan')
        self.hitCompliance = []
        self.sourceMode = self._getSourceMode()
        self.sourceLimits = 100 #Dummy number to be replaced by setSourceRange function
    @Instrument.addOptionSetter("beepEnable")
    def _setBeepEnable(self, enable):
        self.dev.write(":SYST:BEEP:STAT " + ("ON" if enable else "OFF"))

    @Instrument.addOptionSetter("outputEnable")
    def _setOutputEnable(self, enable):
        self.dev.write(":OUTP " + ("ON" if enable else "OFF"))
    
    @Instrument.addOptionSetter("sourceMode")
    def _setSourceMode(self, mode):
        if mode == "voltage":
            # Source a voltage, sense a current
            self.dev.write("SOUR:FUNC VOLT")
            self.dev.write("SENS:FUNC 'CURR'")
            self.sourceMode = 'voltage'
        elif mode == "current":
            # Source a current, sense a voltage
            self.dev.write("SOUR:FUNC CURR")
            self.dev.write("SENS:FUNC 'VOLT'")
            self.sourceMode = 'current'
        else:
            raise ValueError(
                "\"{}\" is not a valid source mode for the Keithley2401.".format(mode) +
                " The mode must either be \"voltage\" or \"current\""
            )


    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
    
    @Instrument.addOptionGetter("sourceMode")
    def _getSourceMode(self):
        mode = self.dev.query("SOUR:FUNC:MODE?")
        if mode == u"VOLT\n":
            return "voltage"
        elif mode == u"CURR\n":
            return "current"
        else:
            raise RuntimeError("Unknown source mode {}".format(mode))

    @Instrument.addOptionSetter("sourceRange")
    def _setSourceRange(self, sourceRange):  
        mode = self.get("sourceMode", forceCached = False)

        if (mode == "voltage"): 
            if float(sourceRange) in (20,10,1,0.1,0.01,0.001):
                self.dev.write("SOUR:VOLT:RANG "+str(sourceRange))
                self.sourceLimits = 1.3*float(sourceRange)
            else:
                raise ValueError(
                    "\"{}\" is not a valid voltage source range for the Keithley2401.".format(sourceRange) +
                    " Valid voltage ranges are: 20, 10, 1, 0.1, 0.01 and 0.001 Volts and equivalent exponential representations."
                )
        elif (mode == "current"):   
            if float(sourceRange) in (1,0.1,0.01,0.001,1E-4,1E-5,1E-6):
                self.dev.write("SOUR:CURR:RANG "+str(sourceRange))   #Dummy function. Have to ignore false inputs and accept 1E-2 as well as 1e-2 etc.
                self.sourceLimits = 1.3*float(sourceRange)
            else:
                raise ValueError(
                    "\"{}\" is not a valid current source range for the Keithley2401.".format(sourceRange) +
                    " Valid voltage ranges are: 1, 0.1, 0.01, 0.001, 1E-5 and 1E-6 Amps and equivalent exponential representations."
                )
    @Instrument.addOptionGetter("sourceLevel")
    def _getSourceLevel(self):
       # mode = self.get("sourceMode", deprecated
        mode = self.sourceMode
        if (mode == "voltage"): #    
           return(float(self.dev.query("SOUR:VOLT?")))
                
        elif (mode == "current"):  
            return(float(self.dev.query("SOUR:CURR?")))
        
        
    @Instrument.addOptionSetter("sourceLevel") 
    def _setSourceLevel(self, sourceLevel):
        #mode = self.get("sourceMode", deprecated
        mode = self.sourceMode
        if (mode == "voltage"): # when sourcing a voltage we want to set the voltage level =)   
            #Limits = float(self.dev.query("SOUR:VOLT:RANG?")) #Get the set range and thus deduce the limits we can put out. Deprecated
            if (self.sourceLimits > sourceLevel and sourceLevel > -self.sourceLimits):
                self.dev.write("SOUR:VOLT "+str(sourceLevel))
                self.currentSourceSetpoint = float(sourceLevel)
            else:
                raise ValueError(
                    "Specified voltage output of \"{} V\" not possible in the given voltage range.".format(sourceLevel)
                )
                
        elif (mode == "current"): # when sourcing a current we want to set the current output level
            #Limits = float(self.dev.query("SOUR:CURR:RANG?")) #Get the set range and thus deduce the limits we can put out. Deprecated
            if (self.sourceLimits > sourceLevel and sourceLevel > -self.sourceLimits):
                self.dev.write("SOUR:CURR "+str(sourceLevel))
                self.currentSourceSetpoint = float(sourceLevel)
            else:
                raise ValueError(
                    "Specified current output of \"{} A\" not possible in the given current range.".format(sourceLevel)
                )

    @Instrument.addOptionSetter("senseRange")
    def _setSenseRange(self, senseRange):
        mode = self.get("sourceMode", forceCached = False)
        if (mode == "voltage"): # when sourcing a voltage we usually sense current aka set its range
            if float(senseRange) in (1.05E-4,1.05E-5,1.05E-6):   
                self.dev.write("SENS:CURR:RANG "+str(senseRange))
            else:
                raise ValueError(
                    "\"{}\" is not a valid current measurement range for the Keithley2401.".format(senseRange) +
                    " Valid current sensing ranges are: 1.05E-4, 1.05E-5 and 1.05E-6 Amps and equivalent representations."
                )
        elif (mode == "current"): # when sourcing a current we sense voltage 
            if float(senseRange) in (21.00,2.10,0.21):   
                self.dev.write("SENS:VOLT:RANG "+str(senseRange))

            else:
                raise ValueError(
                    "\"{}\" is not a valid voltage measurement range for the Keithley2401.".format(senseRange) +
                    " Valid voltage sensing ranges are: 21.00, 2.10 and 0.21 Volts and equivalent representations."
                )

    @Instrument.addOptionGetter("senseLevel")
    def _getSenseLevel(self):
        tempData = self.dev.query_ascii_values(":READ?")
        measInput = float(tempData[1])/self.scaleFactor 
        measOutput = float(tempData[0])
        
        return measInput#,measOutput]
        


    @Instrument.addOptionSetter("compliance")
    def _setCompliance(self,compliance,currOrVolt=None): #currOrVolt is an optional paramter. you can use it to specify whether you want to set the current or voltage protection. If not give, the function will change the current compliance when in voltage sourcing mode and vice versa.
        mode = currOrVolt
        if mode == None:
            mode = self.get("sourceMode", forceCached = False)
            
        if (mode == "voltage" and 0.001E-6 <= float(compliance) <= 1.05): # when sourcing a voltage we want current compliance
            self.dev.write("SENS:CURR:PROT "+str(compliance))


        elif (mode == "current" and 0.2E-3 <= compliance <= 21): # when sourcing a current we want voltage compliance
            self.dev.write("SENS:VOLT:PROT "+str(compliance))


        else:
            raise ValueError(
                "\"{}\" is not a valid current/voltage protection value (compliance) for the Keithley2401.".format(compliance) +
                " Current compliance must be between 1.05 Amps - 0.001E-6 Amps (including bounds). /n Voltage protection values must be between 21 V and 0.2mV."
            )
            
    @Instrument.addOptionGetter("scaleFactor")
    def _getScaleFactor(self):
        return self.scaleFactor
    
    @Instrument.addOptionSetter("scaleFactor")
    def _setScaleFactor(self,scaleFactor):
        self.scaleFactor = scaleFactor
       
    def close(self):
        self.dev.close()
        
    def goTo(self,target,stepsize= 0.001,delay=0.2):
        currentOutput = self.get('sourceLevel')
        sign = 1 if (target>currentOutput) else -1
        sweepArray = np.arange(currentOutput,target+sign*stepsize,sign*stepsize)
        for point in sweepArray:
            self.set('sourceLevel',point)
            time.sleep(delay)
        self.set('sourceLevel',target)     #This set the unit to the final value, even if the target value does not fit together with the stepsize.
        
        



# -*- coding: utf-8 -*-
"""
Created on Tue 23 Feb 2018

@author: z5168331
"""
import pyvisa as visa
import pyneMeas.Instruments.Instrument as Instrument

@Instrument.enableOptions
class Keithley2000(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    defaultOutput = "sourceLevel"
    defaultInput = "senseLevel"

    def __init__(self, address):
        super(Keithley2000, self).__init__()
        self.dev = visa.ResourceManager().open_resource("GPIB0::"+str(address)+"::INSTR")
        print(self.dev.query("*IDN?")) # Probably should query and check we have the right device
        
        #self.dev.write("*RST")
        # self.dev.write("*CLS")
        self.type ="Keithley2000"  #We cna check each instrument for its type and react accordingly
        self.name = 'myKeithley2000'
        self.scaleFactor = 1.0
        self._setSenseMode(self._getSenseMode())


#    @Instrument.addOptionSetter("beepEnable")
#    def _setBeepEnable(self, enable):
#        self.dev.write(":SYST:BEEP:STAT " + ("ON" if enable else "OFF"))

    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
    
    @Instrument.addOptionSetter("senseMode")
    def _setSenseMode(self,senseMode):
        if senseMode == 'voltage':
            self.dev.write('CONF:VOLT')
        elif senseMode == 'current':
            self.dev.write('CONF:CURR')
    
    @Instrument.addOptionGetter("senseMode")
    def _getSenseMode(self):
        senseMode = str(self.dev.query('FUNC?'))[1:-5]
        if senseMode == 'VOLT':
            return 'voltage'
        elif senseMode == 'CURR':
            return 'current'
    
    @Instrument.addOptionSetter("senseRange")   #Voltage sense ranges: 100mV, 1V, 10V,100V, 1000V
    def _setSenseRange(self, senseRange):
        mode = self.get("senseMode", forceCached = False)
        if (mode == "voltage"): #
            if float(senseRange) in (100E-3,1,10,100,1000):   
                self.dev.write("SENS:VOLT:RANG "+str(senseRange))
            else:
                raise ValueError(
                    "\"{}\" is not a valid voltage measurement range for the Keithley2000.".format(senseRange) +
                    " Valid voltage sensing ranges are: 100E-3,1,10,100,1000 Volts and equivalent representations."
                )
        elif (mode == "current"): #
            if float(senseRange) in (10E-3,100E-3,1,3):   
                self.dev.write("SENS:CURR:RANG "+str(senseRange))

            else:
                raise ValueError(
                    "\"{}\" is not a valid current measurement range for the Keithley200.".format(senseRange) +
                    " Valid current sensing ranges are: 10E-3,100E-3,1,3 Amps and equivalent representations."
                )
#
    @Instrument.addOptionGetter("senseLevel")
    def _getSenseLevel(self):
        tempData = self.dev.query_ascii_values(":READ?")
        if abs(tempData[0])> 1E10:
            res = float('nan')
        else: res = float(tempData[0])/self.scaleFactor 
        return res
        

#    @Instrument.addOptionSetter("compliance")
#    def _setCompliance(self,compliance,currOrVolt=None): #currOrVolt is an optional paramter. you can use it to specify whether you want to set the current or voltage protection. If not give, the function will change the current compliance when in voltage sourcing mode and vice versa.
#        mode = currOrVolt
#        if mode == None:
#            mode = self.get("sourceMode", forceCached = True)
#            
#        if (mode == "voltage" and 0.001E-6 <= float(compliance) <= 1.05): # when sourcing a voltage we want current compliance
#            self.dev.write("SENS:CURR:PROT "+str(compliance))
#
#
#        elif (mode == "current" and 0.2E-3 <= compliance <= 21): # when sourcing a current we want voltage compliance
#            self.dev.write("SENS:VOLT:PROT "+str(compliance))
#
#
#        else:
#            raise ValueError(
#                "\"{}\" is not a valid current/voltage protection value (compliance) for the Keithley2401.".format(compliance) +
#                " Current compliance must be between 1.05 Amps - 0.001E-6 Amps (including bounds). /n Voltage protection values must be between 21 V and 0.2mV."
#            )
#            
    @Instrument.addOptionGetter("scaleFactor")
    def _getScaleFactor(self):
        return self.scaleFactor
    
    @Instrument.addOptionSetter("scaleFactor")
    def _setScaleFactor(self,scaleFactor):
        self.scaleFactor = scaleFactor
       
    def close(self):
        self.dev.close()
        
    def goTo(self,target,stepsize= 0.001,delay=0.2):
        return
        



# -*- coding: utf-8 -*-
"""
Jakob Seidl, Nanoelectronics Group UNSW Sydney
"""
import pyvisa as visa
import pyneMeas.Instruments.Instrument as Instrument
import time
import math
import numpy as np




@Instrument.enableOptions
class SRS830(Instrument.Instrument):
    """
    SRS830 Lock-In Amplifier
    >>> myLockin = I.SRS830(GPIBAddress)
    """
    defaultOutput = "sourceLevel"
    defaultInput = "senseLevel"
    def __init__(self, address):
        super(SRS830,self).__init__()
        self.dev = visa.ResourceManager().open_resource("GPIB0::"+str(address)+"::INSTR")
        print(self.dev.query("*IDN?")) # Probably should query and check we have the right device
        
        self.outputMode = "amplitude" # or "frequency"
        self.sweepDelay = 0.5
        self.type ="SRS830"
        self.name = 'mySRS830'
        self.scaleFactor = 1.0
        self.input = self._getInput()
        #AutoRange:
            
        self.autoSensitivityEnable = False
        self.possibleCurrRanges = [9,12,15,18,21,24,26]
        self.rangeIndex =  6 #random

        self.rangeDic = {  #we shift the lower intervals 30% (-> 0.7*number) downwards so they dont overlap with the bounds when we shift up.
            0:[15E-18,1.5E-12],
            1:[0.7*1.5E-12,15E-12],
            2:[0.7*15E-12,150E-12],
            3:[0.7*150E-12,1.5E-9],
            4:[0.7*1.5E-9,15E-9],
            5:[0.7*15E-9,150E-9],
            6:[0.7*150E-9,1.5E-6], 
            }
    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
    
    @Instrument.addOptionSetter("frequency")
    def _setFrequency(self,frequency):
        self.dev.write("FREQ  "+str(frequency))

    @Instrument.addOptionGetter("frequency")
    def _getFrequency(self):
        return float(self.dev.query("FREQ?"))
    
    
    @Instrument.addOptionSetter("amplitude")
    def _setAmplitude(self,amplitude):
        self.dev.write("SLVL  "+str(amplitude))

    @Instrument.addOptionGetter("amplitude")
    def _getAmplitude(self):
        return float(self.dev.query("SLVL?"))

    
    @Instrument.addOptionGetter("senseLevel")
    def readXY(self): # This function is a bit more ellaborate since snap return a string comprised of both X and Y values. The function then separates this string, i.e. looks for the separation coma and both exponents... 
        XY = self.dev.query("SNAP?1,2") # ONE could also use OUTP? i (i=1,2,3,4 X,Y,R theta)
        divideIndex = XY.find(",")
        X = XY[0:divideIndex]
        #print(X)
        Y = XY[divideIndex+1:-1]
        #print(Y)
        #Voltage measurements:
        if self.input == 'A' or self.input =='A-B': #
            XexpIndex = X.find("e-")
            YexpIndex = Y.find("e-")
            if (XexpIndex == -1 or YexpIndex == -1):
                return [float(X)*(1/self.scaleFactor),float(Y)*(1/self.scaleFactor)]
            else:
                XList =list(X)
                if XList[XexpIndex+2] =="0":
                    XList[XexpIndex+2] =""
                Xfinal = float("".join(XList))
                YList =list(Y)
                if YList[YexpIndex+2] =="0":
                    YList[YexpIndex+2] =""
                Yfinal = float("".join(YList))
                return [(1/self.scaleFactor)*Xfinal,(1/self.scaleFactor)*Yfinal]
        #Current measurements          
        XexpIndex = X.find("e-")
        if XexpIndex != -1:
            # print("Xexp index is "+str(XexpIndex))
           XList =list(X)
           if XList[XexpIndex+2] =="0":
              XList[XexpIndex+2] =""
           Xfinal = float("".join(XList))
        else:
            Xfinal = 0 # we need to do something like this because if the result is already zero or some odd/weird number, the exponent finding process above breaks down
        
        YexpIndex = Y.find("e-")    
        if YexpIndex != -1:
            #  print("Yexp index is "+str(YexpIndex))
           YList =list(Y)
           if YList[YexpIndex+2] =="0":
              YList[YexpIndex+2] =""
           Yfinal = float("".join(YList))
        else:
           Yfinal = 0
           
        if self.autoSensitivityEnable:
            self._autoSensitivity([Xfinal,Yfinal])
        return [(1/self.scaleFactor)*Xfinal,(1/self.scaleFactor)*Yfinal]
    

    def _autoSensitivity(self,givenValue):
            R = math.sqrt(givenValue[0]**2+givenValue[1]**2)
#            print('#############\n')
#            print('current value is'+str(R)+'\n')
#            print('current range index is: '+str(self.rangeIndex)+'\n')
#            print('the boundaries here are: '+str(self.rangeDic[self.rangeIndex][0]) +'and' + str(self.rangeDic[self.rangeIndex][1])+'\n')
#            print('#############\n')
            if (self.rangeDic[self.rangeIndex][0] >= R): #and not math.isnan(R)
                a = self.rangeIndex -1
                if a in range(7):
                    self.rangeIndex = a
                   # print(self.rangeIndex)
                    self.dev.write("SENS "+str(self.possibleCurrRanges[self.rangeIndex]))
                    time.sleep(2)
            elif (R > self.rangeDic[self.rangeIndex][1]):
                a = self.rangeIndex +1
                if a in range(7):
                    self.rangeIndex = a
                    #print(self.rangeIndex)
                    self.dev.write("SENS "+str(self.possibleCurrRanges[self.rangeIndex]))
                    time.sleep(2)
            else:
                pass

    @Instrument.addOptionSetter("sourceLevel")
    def _setSourceLevel(self,sourceLevel):
        self.dev.write(("SLVL " if self.outputMode == "amplitude" else "FREQ ") +str(sourceLevel))
        time.sleep(self.sweepDelay) 
    @Instrument.addOptionGetter("sourceLevel")
    def _getSourceLevel(self):
         return float(self.dev.query(("SLVL?" if self.outputMode == "amplitude" else "FREQ?")))

    
    @Instrument.addOptionSetter("timeConst")    
    def _setTimeConst(self,timeConst): #Index can be between 0 (10 microsec to 19 30ksec)
        possibleConstants = [10E-6,30E-6,100E-6,300E-6,1E-3,3E-3,10E-3,30E-3,100E-3,300E-3,1,3,10,30,100,300]
        if float(timeConst) in possibleConstants: #Not including the highest kilosecond integration times since not never used.
            self.dev.write("OFLT "+str(possibleConstants.index(timeConst)))
        else:
            raise ValueError(
                            "\"{}\" is not a valid time Constant for the SRS830.".format(timeConst) +
                            " Valid time constants are: (10,30,100,300)X 1E-6, (1,3,10,30,100,300)X 1E-3 and (1,3,10,30,100,300)X 1E-0 seconds."
                            )
    
    @Instrument.addOptionSetter("senseRange")        
    def _setSenseRange(self,senseRange):
        possibleSenseRanges = [1E-9,2E-9,5E-9,10E-9,20E-9,50E-9,100E-9,200E-9,500E-9,1E-6,2E-6,5E-6,10E-6,20E-6,50E-6,100E-6,200E-6,500E-6,1E-3,2E-3,5E-3,10E-3,20E-3,50E-3,100E-3,200E-3,500E-3,1] #For voltages !!
        if senseRange in possibleSenseRanges:
            sensIndex = possibleSenseRanges.index(senseRange)-1
            if self.rangeIndex != sensIndex: #Only change rhe range if it is different than before. This will be helpfull with the auoRange function True
                self.dev.write("SENS "+str(sensIndex))
                time.sleep(0.4)
                #self.rangeIndex = sensIndex
            else:
                pass
        else:
            raise ValueError(
                            "\"{}\" is not a valid senseRange for the SRS830.".format(senseRange) +
                            " Valid sensitivities are:\"{}\"  Volts or microAmps respectively.".format(possibleSenseRanges)
                            )
    @Instrument.addOptionGetter("senseRange")
    def _getSenseRange(self):
        possibleSenseRanges = [1E-9,2E-9,5E-9,10E-9,20E-9,50E-9,100E-9,200E-9,500E-9,1E-6,2E-6,5E-6,10E-6,20E-6,50E-6,100E-6,200E-6,500E-6,1E-3,2E-3,5E-3,10E-3,20E-3,50E-3,100E-3,200E-3,500E-3,1] #For voltages !!
        tempRes = int(self.dev.query('SENS?'))
        return(possibleSenseRanges[tempRes])
    
    
    @Instrument.addOptionSetter("input")
    def _setInput(self,inputSetting):  #I1 = I(1E6) and I2 = I(1E8)
        possibleSettings = ["A","A-B","I1","I2"]
        if str(inputSetting) in possibleSettings:
            self.dev.write("ISRC "+str(possibleSettings.index(inputSetting)))
            self.input = str(inputSetting)
    
    @Instrument.addOptionGetter("input")
    def _getInput(self):
        possibleSettings = ["A","A-B","I1","I2"]
        modeIndex = int(self.dev.query("ISRC?"))
        return possibleSettings[modeIndex]
    
    @Instrument.addOptionGetter("scaleFactor")
    def _getScaleFactor(self):
        return self.scaleFactor
    
    @Instrument.addOptionSetter("scaleFactor")
    def _setScaleFactor(self,scaleFactor):
        self.scaleFactor = scaleFactor
    
    @Instrument.addOptionSetter("sweepParameter")
    def _setSweepParameter(self,sweepParameter):
        if sweepParameter in ("frequency","amplitude"): 
            self.outputMode = str(sweepParameter)
        else:
            raise ValueError(
                            "\"{}\" is not a valid sweepParameter for the SRS830.".format(sweepParameter) +
                            " You can either sweep the reference 'frequency' or the sine output 'amplitude'."
                            )
    
    def _getSenseIndex(self):
        return int(self.dev.query("SENS?"))
    
    @Instrument.addOptionGetter("sweepParameter")
    def _getSweepParameter(self):
        return self.outputMode
    
    @Instrument.addOptionSetter('sweepDelay')
    def _setSweepDelay(self,delay):
        self.sweepDelay = delay
    @Instrument.addOptionGetter('sweepDelay')
    def _getSweepDelay(self):
        return self.sweepDelay
    
    @Instrument.addOptionSetter("autoSensitivity")
    def _setAutoSensitivity(self,enable):
        self.autoSensitivityEnable = True if enable else False
    @Instrument.addOptionGetter("autoSensitivity")
    def _getAutoSensitivity(self):
        return self.autoSensitivityEnable
    
    def autoPhase(self):
        self.dev.write("APHS")
        time.sleep(3)
    @Instrument.addOptionSetter("phase")
    def _setPhase(self,phase):
        self.dev.write("PHAS "+str(phase))
        
    @Instrument.addOptionGetter("phase")
    def _getPhase(self):
        return float(self.dev.query("PHAS?"))
    
    def close(self):
        self.dev.close()
        
    def goTo(self,target,stepsize= 0.01,delay=0.2):
        currentOutput = self.get('sourceLevel')
        sign = 1 if (target>currentOutput) else -1
        sweepArray = np.arange(currentOutput,target+sign*stepsize,sign*stepsize)
        for point in sweepArray:
            self.set('sourceLevel',point)
            time.sleep(delay)
        self.set('sourceLevel',target)



from pyneMeas.Instruments.TimeMeas import TimeMeas
from pyneMeas.Instruments.USB6216In import USB6216In
from pyneMeas.Instruments.USB6216Out import USB6216Out
from pyneMeas.Instruments.Keithley2401 import Keithley2401
from pyneMeas.Instruments.YokogawaGS200 import YokogawaGS200
from pyneMeas.Instruments.Keithley2000 import Keithley2000
from pyneMeas.Instruments.SRS830 import SRS830

from pyneMeas.Instruments.NoiseGenerator import LinearNoiseGenerator
from pyneMeas.Instruments.NoiseGenerator import SineNoiseGenerator
from pyneMeas.Instruments.NoiseGenerator import DoubleNoiseGenerator





"""
VERSION 4.0
@author: Jakob Seidl
jakob.seidl@nanoelectronics.physics.unsw.edu.au
"""
from collections import Iterable
import json
from pyneMeas.utility import GlobalMeasID as ID
import pyneMeas.Instruments.Instrument as Instrument
import numpy as np
import scipy.io as sio
import time
import os
import pandas as pd
from itertools import product
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import warnings


# Specify plotting options:
#Plot style presets: e.g, 'fivethirtyeight' or 'ggplot' or 'seaborn. More examples -> https://matplotlib.org/3.3.3/gallery/style_sheets/style_sheets_reference.html
plt.style.use('ggplot')
# mpl.rcParams['axes.linewidth'] = 1.0    #  boxplot linewidth
# mpl.rcParams.update({'font.size': 11})  #  fontsize
########################################



# mpl.use('TkAgg') #This should be the default on windows I think, but on my mac it isnt



# For each input point:
# 1. Sets the instruments to the input point by calling the inputSetters with the point
# 2. Calls each outputReaders to get the output point
# 3. Calls outputReceiver with the point
# After all input points have been processed, outputFinisher is called
# 

def sweepAndSave(
                 inputDict,
                 extraInstruments = [],
                 saveEnable = True,
                 delay = 0.0,
                 plotVars = None,
                 # plotString  = ['go--','ro-','bo--','o-'], #linePlot: symbol (-; -- or o,s,^,>   plus a colour: 'r' red, 'g' green etc.
                 comments = "No comments provided by user",
                 plotParams = [('go--','linear-linear'),('ro-','linear-linear')]*10,
                 saveCounter = 10,
                 breakCondition = None,
                 plotCounter = 1,
                 plotSize = (10,10),
                 plotAlpha = 0.8,
                    ):

    """sweepAndSave is the main function of this module and THE sweepfunction. It is the only function that the user actually calls himlself
    - all other functions provided here all called 'under the hood by sweepAndsave'. It can be subdivided into three parts:
    (1): Initialize: Check if user input is valid, create data and log files on disk and set up the plot object. All this is done ONCE ONLY.
    (2): sweep() function which calls receiver() within: LOOP Section: This is the actual sweep,
    i.e., iterative process carried out over every single datapoint in the inputArray. This is: (I) set setter Instr to inputArray point, then Query all meas instruments,
    (II) append data to files and plots and (III) every N-th iteration (default =10), force the file to be written on disk.
    (3): Wrap-up: Write the final instrument parameters and data to file, save the plot as .png, close all file connections (good practice!)"""
    # def unpackInputDict(inputDict):
    #     keyList = ['basePath','fileName','inputHeaders','sweepArray',
    #                'inputSetters','outputHeaders','outputReaders']
    #     return [inputDict[key] for key in keyList]
    def unpackInputDict(inputDict):
        keyList = ['basePath','fileName','setters','sweepArray',
                   'readers']
        return [inputDict[key] for key in keyList]

    def instrumentsHeadersfromDict(inputDict):
        instruments, headers = [], []
        for key, item in inputDict.items():
            instruments.append(key)
            if isinstance(item, list) or isinstance(item, tuple):
                for i in item:
                    headers.append(i)
            else:
                headers.append(item)
        return (instruments, headers)

    inputDict = checkInputDict(inputDict,saveEnable)
    # basePath,fileName,inputHeaders,inputPoints, inputSetters,outputHeaders, outputReaders = unpackInputDict(inputDict)
    basePath, fileName, setters, inputPoints, readers = unpackInputDict(inputDict)
    inputSetters,inputHeaders = instrumentsHeadersfromDict(setters)
    outputReaders,outputHeaders = instrumentsHeadersfromDict(readers)

    lenInputList = len(inputPoints)

    ######################################################################## Part(1)
    #########################################################################################################################################################
    #Turn input array into itertools.product if it isnt already. Since our sweepArray is usually a 1D array anyway, this is usually not necessary and is more of a relic than a feature:
    if (type(inputPoints) == product):
        pass
    elif (type(inputPoints) == list or type(inputPoints) == np.ndarray):
        inputPoints = product(inputPoints)
    else:
        pass


    # Check if the plotting parameters ('plotVars') exist in in- and outputHeaders:
    # checkPlotHeaders(inputHeaders,outputHeaders,plotVars)


    inputHeaders = flatten(inputHeaders);\
    outputHeaders = flatten(outputHeaders); #Make sure we have simple lists, not lists within lists etc..
    # allHeaders = inputHeaders + outputHeaders

    ### New breakCond Stuff ####
    if breakCondition != None and len(breakCondition) == 3:
        breakIndex =  outputHeaders.index(breakCondition[0])
    else: breakIndex = None

    # Prepare a dict for the data too. This dict will be used to write data to a .mat file which can be conveniently read by Matlab or Python
    pointsDict = {}
    for header in flatten((flatten(inputHeaders), flatten(outputHeaders))):
        pointsDict[header] = []

    ID.increaseID()
    measurementName = str(ID.readCurrentSetup()) + str(ID.readCurrentID())
    startTime = time.time()  # Start time

    # Main loop if user wants to save data to file:
    if(saveEnable):

        fileName = str(ID.readCurrentSetup()) + str(ID.readCurrentID()) + "_" + fileName




        # Make a copy of the initial configuration of the instruments
        # instruments = set(filter(  # Old donny version, let's see when we run into problems with the simple one below!
        #     lambda i: issubclass(type(i), Instrument.Instrument),
        #     inputSetters + outputReaders + extraInstruments
        # ))
        instruments  = set(inputSetters + outputReaders + extraInstruments)

        config = {}
        for instrument in instruments: #This goes through the list of all instruments and queries all options that have a associated 'get()' method. E.g., 'sourceMode' for the Keithely2401
            config["{}-{}-{}".format(instrument.get('name'),type(instrument).__name__,len(config)+1)] = instrument.getOptions()#The ['key'] for each instrument is its 'name' and its type.


        #  write the initial config to the LOG file:
        log = open(basePath+fileName +"_LOG"+ ".tsv", "w")
        log.write("Measurement Log file for measurement >>> "+ str(ID.readCurrentSetup()) + str(ID.readCurrentID())+" <<< \n")
        log.write("Starting time and date: "+time.asctime(time.localtime(time.time()))+"\n")
        log.write("\n")
        log.write("Set/sweeped variables/instruments:\n")
        log.write("Variable\tInst name\tInst type\n")
        log.write("-----------------------------------\n")
        for header,name, Insttype in zip(inputHeaders, [setter.name for setter in inputSetters], [setter.type for setter in inputSetters]):
            log.write(f"{header}\t{name}\t{Insttype}\n")

        log.write("\n")
        log.write("Measuring variables/instruments:\n")
        log.write("Variable\tInst name\tInst type\n")
        log.write("-----------------------------------\n")
        for header,name, Insttype in zip(outputHeaders, [getter.name for getter in outputReaders], [getter.type for getter in outputReaders]):
            log.write(f"{header}\t{name}\t{Insttype}\n")


        log.write("\n")
        log.write("User comments: "+str(comments) +"\n")
        log.write("-----------------------------------\n")
        log.write("Delay = "+str(delay)+"s \n")
        log.write("Initial instrument configuration\n")
        log.write("-----------------------------------\n")
        log.write(json.dumps(config, indent = 4, sort_keys = True)) #Writes all initial instrument paramters in intented Json format
        log.write("\n-----------------------------------\n")
        log.close()


        #Write data headers to plain text file :
        tsv = open(basePath + fileName + ".tsv", "w")
        tsv.write("\t".join(flatten(inputHeaders))+ "\t")
        tsv.write("\t".join(flatten(outputHeaders)) + "\n")


        ##############          Prepare Plotting: ###############

        # measurementName = str(ID.readCurrentSetup()) + str(ID.readCurrentID()) # returns e.g. At104

        Xvalues1 = [];Yvalues1 = [];Xvalues2 = [];Yvalues2 = [] #Generate empty lists of X and Y Data. This is used later in the plotting routine.
        if plotVars is not None:
            ############## Initialize the plot. Actual plotting happens within receiver() within save() ###############
            plt.ion()  # this is the call to matplotlib that allows dynamic plotting
            fig = plt.figure(figsize = plotSize)

            lineObjs = []
            axObjs = []
            nSubplot = len(plotVars)

            # for i in range(1,nSubplot+1):
            nRows = int(np.ceil(np.sqrt(nSubplot)))
            nCols = int(np.floor(np.sqrt(nSubplot)))
            if nRows*nCols < nSubplot:
                nRows += 1

            for index,(xyAxisName,plotParamTuple) in enumerate(zip(plotVars,plotParams)):
                xAxisName,yAxisName = xyAxisName
                plotString, axisString = plotParamTuple
                subplotIndex = nRows*100+nCols*10+index+1
                axObj = fig.add_subplot(subplotIndex)

                # create a variable for the line so we can later update it (in the receiver)
                line, = axObj.plot(0.01, 0.02, plotString, alpha = plotAlpha)

                xScale = axisString.split('-')[0]
                yScale = axisString.split('-')[1]
                axObj.set_xscale(xScale)
                axObj.set_yscale(yScale)

                # update plot label/title
                if plotVars !=None and len(plotVars) == nSubplot:
                    plt.ylabel(yAxisName)
                    plt.xlabel(xAxisName)
                if index == 0:
                    plt.title(f'{measurementName}')

                lineObjs.append(line)
                axObjs.append(axObj)
            plt.subplots_adjust(left=0.125, right=0.94, bottom=0.1, top=0.93, wspace=0.5, hspace=0.5)
            plt.show()
        ############### END initialize plot ###############################




        ######################################################################## Part(2) --The loop-- ########################################################################
        #########################################################################################################################################################


        ########  The receiver() function is a sub-section of the save() function and is called for EACH point in the sweepArray.
        #It does two things: 1) Append the set of measured values (e.g. results from you SRS830 and K2401) to your measurement text file (.tsv)
            #and append it to your python dictionary of results (used for export as .mat file in the end). 2) It updates the current plot with the new data.

        ##############  Definition of receiver() ###############################
        def receiver(inputPoint, outputPoint,counter,lenInputList,forceSave = False, forcePlot = False,):

            # checkPointMatchesHeaders(inputPoint, inputHeaders)
            # checkPointMatchesHeaders(outputPoint, outputHeaders)

            for value, header in zip(flatten(inputPoint), flatten(inputHeaders)):
                pointsDict[header].append(value)
            for value, header in zip(flatten(outputPoint), flatten(outputHeaders)):
                pointsDict[header].append(value)

            tsv.write("\t".join(map(str, flatten(inputPoint))) + "\t") #takes the input points, 'flattens' the list (aka gets rid of unecessary lists in lists) turns them into strings and writes them separated by a tab \t in the tsv file.
            tsv.write("\t".join(map(str, flatten(outputPoint))) + "\n")

            progressBar(counter, lenInputList, barLength=20)
            #these force saving commands should probably only be executed every tenth iteration or so to speed things up.
            if (counter%saveCounter == 0 or forceSave):
                tsv.flush()   #These two commands force the tsv file and .mat file to be saved to disk. Otherwise the file will be lost when killing the program
                os.fsync(tsv.fileno())
                sio.savemat(basePath +fileName + '.mat', pointsDict)

            if plotVars is not None:
                xyDatSet = [(pointsDict[p1],pointsDict[p2]) for p1,p2 in plotVars]
                #Do the actual Plotting:
                if (counter % plotCounter == 0 or forcePlot):
                    for lineOb, axObj, xyTuple in zip(lineObjs, axObjs, xyDatSet):
                        x, y = xyTuple
                        lineOb.set_ydata(y)
                        lineOb.set_xdata(x)

                        stdY = np.std(y)
                        stdX = np.std(x) / 4
                        try: #Introduced this since sometimes 'NaNs' or other chunk data may impede setting the axis limits properly
                            with warnings.catch_warnings():  #used to suppress the annoying matplotlib warning about singular axe
                                warnings.simplefilter("ignore")
                                axObj.set_ylim([np.min(y) - stdY, np.max(y) + stdY])
                                axObj.set_xlim([np.min(x) - stdX, np.max(x) + stdX])
                        except:
                            pass

                    plt.pause(0.00001)
                ############## END Definition of receiver() ###############################

        #sweep() does the actual sweep and calls receiver() defined just above! sweep() is defined just below, outside of the sweepAndSave() definition

        sweep(inputPoints, inputSetters, outputReaders, receiver,delay,breakCondition,breakIndex,lenInputList)






        ######################################################################## Part(3) ########################################################################
        #########################################################################################################################################################
        ###### Wrapping up the measurement: close the data.tsv file, write the final settings of all instruments in .log file,
        #save a final version of the data to .mat format and save the figure created as .png

        tsv.flush()  # These three commands force the tsv file and .mat file to be saved to disk. Otherwise the file will be lost when killing the program
        os.fsync(tsv.fileno())
        tsv.close()
        sio.savemat(basePath +fileName + '.mat', pointsDict)

        elapsedTime = time.time()-startTime
        log = open(basePath +fileName +"_LOG"+ ".tsv", "a")
        log.write("Ending time and date: "+time.asctime(time.localtime(time.time()))+"\n")
        log.write("Time elapsed: "+str(elapsedTime)+" seconds." +"\n")
        log.write("Final instrument configuration: \n")
        log.write("-----------------------------------\n")
        log.write(json.dumps(config, indent = 4, sort_keys = True)) #Writes all the instrument parameters in indeted json format
        log.write("\n-----------------------------------\n")
        log.close()

        if plotVars is not None:
            plt.savefig(basePath +fileName+'.png') #Save Plot as .png as additional feature (only if plotting parameters were specified
            plt.close('all')







    # Main loop if user doesnt want to save data to file:
    elif(not saveEnable): #


        def receiverNoSave(inputPoint, outputPoint):
            for value, header in zip(flatten(inputPoint), flatten(inputHeaders)):
                pointsDict[header].append(value)
            for value, header in zip(flatten(outputPoint), flatten(outputHeaders)):
                pointsDict[header].append(value)

        sweepNoSave(inputPoints, inputSetters, outputReaders,receiverNoSave , delay,breakCondition,breakIndex,lenInputList)  # This does the actual sweep (without saving)!





    # Wrap Up:
    elapsedTime = time.time()-startTime
    print(f'/>>>>>> Finished measurement {measurementName:} | Duration: {elapsedTime:.1f} seconds = {elapsedTime/60:.1f} min  <<<<<<<')
    return pd.DataFrame.from_dict(pointsDict)


######################################################################## END of sweepAndSave() ########################################################################
#######################################################################################################################################################################

# def receiverNoSave(inputPoint, outputPoint, outputHeaders):
#     for value, header in zip(flatten(inputPoint), flatten(inputHeaders)):
#         pointsDict[header].append(value)
#     for value, header in zip(flatten(outputPoint), flatten(outputHeaders)):
#         pointsDict[header].append(value)






##################################################################
def sweep(inputPoints, inputSetters, outputReaders, receiver,delay,breakCondition,breakIndex,lenInputList):
    """sweep() defines the 'actual sweep',i.e.,, we define what is done for each 'inputPoint' of the array we want to sweep over. """
    prevPoint = None
    counter = 0

    def checkPointBreaks():
        if breakCondition[1] == '>':
            return (flatten(outputPoint)[breakIndex] < breakCondition[2])
        elif breakCondition[1] == '<':
            return (flatten(outputPoint)[breakIndex] > breakCondition[2])

    # Actual loop over all points in inputArray:
    for inputPoint in inputPoints:
#            while(running): #If we wanted something like a break condition, this would be the place to put them. Not yet implemented, though.
                if len(inputPoint) != len(inputSetters):#Usually len(inputPoint) is 1 since it's a single point in a 1D array. One instrument, one value.
                    raise ValueError("Length of input point does not match length of input setters")

                #We define the 'previous' state of the sweep so we are able to
                if prevPoint == None: #For the first point of each sweep, this is the case. The setter.goTo(target) then slowly goes to the first value. This is mainly to catch user errors.
                    prevPoint = [None] * len(inputPoint) #Usually len(inputPoint) is 1 since it's a single point in a 1D array.
                    for value,setter in zip(inputPoint,inputSetters):
                        setter.goTo(value)
                #A Set source instrument
                for value, prevValue, setter in zip(inputPoint, prevPoint, inputSetters):
                    # Avoid setting a value if it is already set to it
                    if value != prevValue: #only change outputs if they are in fact different to the previous once. Saves time.
                        if callable(setter):# In principle one could directly pass a (lambda) function instead of a real instrument. Then this block is carried out.
                            setter(value)
                        else: # This is carried out when a real Instrument is passed to the SweepAndSave function, so nearly always.
                            setter.set(type(setter).defaultOutput, value)

                prevPoint = inputPoint
                #### B Reading out all instruments defined as 'outputReaders'
                time.sleep(delay) #This is the delay specified by the user, typicall 0.2s
                outputPoint = []
                for reader in outputReaders:
                    if callable(reader): # In principle one could directly pass a (lambda) function instead of a real instrument. Then this block is carried out.
                        tempRes = reader()

                        outputPoint.append(tempRes)

                    else: #However, usually we provide a 'real' instrument object and the appropriate instrument.get('readVariable') is called.
                        tempRes = reader.get(type(reader).defaultInput)
                        outputPoint.append(tempRes)

                #  !!!! BREAK !!!!!!!! BREAK !!!!!!!! BREAK !!!!
                if (breakIndex == None or checkPointBreaks()):
                    receiver(inputPoint, outputPoint,counter,lenInputList)
                    counter = counter+1
                else:
                    print(f"Terminated measurement: Reached breakCondition: \nMeasured variable: {breakCondition[0]} reached {flatten(outputPoint)[breakIndex]} which is {breakCondition[1]} {breakCondition[2]}")
                    break
    receiver(inputPoint, outputPoint, counter,lenInputList, forceSave=True, forcePlot=True,)




#               inputPoints, inputSetters, outputReaders, receiver,delay,breakCondition,breakIndex,lenInputList
def sweepNoSave(inputPoints, inputSetters, outputReaders,receiverNoSave,delay,breakCondition,breakIndex,lenInputList): #Since by default the 'saveEnable' option is True, this funciton is barely ever called.

    def checkPointBreaks():
        if breakCondition[1] == '>':
            return (flatten(outputPoint)[breakIndex] < breakCondition[2])
        elif breakCondition[1] == '<':
            return (flatten(outputPoint)[breakIndex] > breakCondition[2])

    prevPoint = None
    for inputPoint in inputPoints:
        if len(inputPoint) != len(inputSetters):
            raise ValueError("Length of input point does not match length of input setters")

        #We define the 'previous' state of the sweep so we are able to only change outputs if they are in fact different to the previous once. Saves time.
        if prevPoint == None: #For the first point of each sweep, this is the case.
            prevPoint = [None] * len(inputPoint) #Usually len(inputPoint) is 1 since it's a single point in a 1D array.

        for value, prevValue, setter in zip(inputPoint, prevPoint, inputSetters):
            # Avoid setting a value if it is already set to it
            if value != prevValue:
                if callable(setter):
                    setter(value)
                else:
                    setter.set(type(setter).defaultOutput, value)
        prevPoint = inputPoint

        time.sleep(delay)
        outputPoint = []
        for reader in outputReaders:
            if callable(reader):
                outputPoint.append(reader())
            else:
                outputPoint.append(reader.get(type(reader).defaultInput))

        if (breakIndex == None or checkPointBreaks()):
            receiverNoSave(inputPoint, outputPoint)
        else:
            print(
                f"Terminated measurement: Reached breakCondition: \nMeasured variable: {breakCondition[0]} reached {flatten(outputPoint)[breakIndex]} which is {breakCondition[1]} {breakCondition[2]}")
            break
    receiverNoSave(inputPoint, outputPoint)
#################################### From here on unimportant and helper functions ################
###################################################################################################

#Helper functions from here on:

# def checkInputDict(inputDict):
#     requiredKeys = {'basePath','fileName','inputHeaders',
#                     'sweepArray','inputSetters','outputHeaders','outputReaders'}
#     inputArgs = set(inputDict.keys())
#
#
#     # 5. inputHeaders
#     if isinstance(inputDict['inputHeaders'],str):
#         if not ((" " in inputDict['inputHeaders']) or (inputDict['inputHeaders'][0].isdigit())):
#             inputDict['inputHeaders'] = [inputDict['inputHeaders']]
#         else: raise ValueError(f"inputHeaders attribute must be one string, or list of strings. NO empty spaces or leading digits are allowed to preserve Matlab compatibility.\nYour entered '{inputDict['inputHeaders']}'.")
#     elif isinstance(inputDict['inputHeaders'],list):
#         for header in inputDict['inputHeaders']:
#             if (not isinstance(header,str) or (" " in header.strip()) or (header[0].isdigit()) ):
#                 raise ValueError(f"inputHeaders attribute must be one string, or list of strings. NO empty spaces or leading digits are allowed to preserve Matlab compatibility.\nYour entered {inputDict['inputHeaders']}.")
#     else: raise ValueError(f"inputHeaders attribute must be one string, or list of strings. You entered: {inputDict['inputHeaders']} of type: {type(inputDict['inputHeaders']).__name__}")

#
#     # 7. inputSetters
#     if type(inputDict['inputSetters']) != list:
#         if 'Instruments' not in str(type(inputDict['inputSetters'])):
#             raise ValueError(f"inputSetters attribute needs to be single instrument object or list of instrument objects (when more than one setter is defined).\nYour entered: {inputDict['inputSetters']}")
#         else: inputDict['inputSetters'] = [inputDict['inputSetters']]
#     else:
#         for listElement in inputDict['inputSetters']:
#             if 'Instruments' not in str(type(listElement)):
#                 raise ValueError(
#                     f"inputSetters attribute needs to be single instrument object or list of instrument objects (when more than one setter is defined).\nYou entered {listElement}, which is of type {type(listElement).__name__}")
#
#
#     # 8. outputHeaders
#     if isinstance(inputDict['outputHeaders'],str):
#         if not ((" " in inputDict['outputHeaders']) or (inputDict['outputHeaders'][0].isdigit())):
#             inputDict['outputHeaders'] = [inputDict['outputHeaders']]
#         else: raise ValueError(f"outputHeaders attribute must be one string, or list of strings. NO empty spaces or leading digits are allowed to preserve Matlab compatibility.\nYour entered '{inputDict['outputHeaders']}'.")
#     elif isinstance(inputDict['outputHeaders'],list):
#         for header in inputDict['outputHeaders']:
#             if (not isinstance(header,str) or (" " in header.strip()) or (header[0].isdigit()) ):
#                 raise ValueError(f"outputHeaders attribute must be one string, or list of strings. NO empty spaces or leading digits are allowed to preserve Matlab compatibility.\nYour entered {inputDict['outputHeaders']}.")
#     else: raise ValueError(f"outputHeaders attribute must be one string, or list of strings. You entered: {inputDict['outputHeaders']} of type: {type(inputDict['outputHeaders']).__name__}")
#
#     # 9. outputReaders
#     if type(inputDict['outputReaders']) != list:
#         if 'Instruments' not in str(type(inputDict['outputReaders'])):
#             raise ValueError(f"outputReaders attribute needs to be single instrument object or list of instrument objects (when more than one setter is defined).\nYour entered: {inputDict['outputReaders']}")
#         else: inputDict['outputReaders'] = [inputDict['outputReaders']]
#     else:
#         for listElement in inputDict['outputReaders']:
#             if 'Instruments' not in str(type(listElement)):
#                 raise ValueError(
#                     f"outputReaders attribute needs to be single instrument object or list of instrument objects (when more than one setter is defined).\nYou entered {listElement}, which is of type {type(listElement).__name__}")
#
#     # 10. Make sure
#
#     return inputDict


def checkInputDict(inputDict,saveEnable):
    requiredKeys = {'basePath','fileName','setters',
                    'sweepArray','readers'}
    inputArgs = set(inputDict.keys())

    missingInputs = requiredKeys.difference(inputArgs)
    unusedInputs = inputArgs.difference(requiredKeys)

    # 1. Assert we have all essential input keys
    assert len(missingInputs) == 0, f"\n Please define the following MISSING ARGUMENTS in your input dictionary: \n {[key for key in missingInputs]}"

    # 2. Letting user know some of his inputs will be ignored. Happens when extra dict keys are used.
    if len(unusedInputs) !=0:
        warnings.warn(f"Warning: The following arguments in your input dictionary will be ignored: \n {[arg for arg in unusedInputs]} ")

    #  3. basePath
    assert isinstance(inputDict['basePath'],str), f"The 'basePath' attribute of the input dictionary needs to be a string! You entered {inputDict['basePath']}."
    if os.path.exists(inputDict['basePath']):
        if saveEnable: print(f'Saving data to existing directory: {inputDict["basePath"]} (if saving is enabled)')
        else: print('Data not saved to disk. Use pandas.dataFrame return value of sweep() to use data.')
    else:
        os.mkdir(inputDict['basePath'])
        print(f'Created data storage directory: {inputDict["basePath"]}')
    if inputDict['basePath'][-1] != '/': # When the user specifies a path not ending on /, we still want to save on the lowest subfolder. -> Add /
        inputDict['basePath'] = inputDict['basePath'] + '/'

    #  4. fileName
    assert isinstance(inputDict['fileName'],str), f"The 'fileName' attribute of the input dictionary needs to be a string! You entered {inputDict['fileName']}."

    # 5. setters
    if not isinstance(inputDict['setters'],dict):
        raise  ValueError(f'Your input dictionary with key "setters" mus be a dictionary itself. You assigned: {inputDict["setters"]} of type: {type(inputDict["setters"]).__name__}\n Please assign the instrument as "key" and the corresponding variable name(s) as values')

    # 6. sweepArray
    assert type(inputDict['sweepArray']) in [range,list,np.ndarray],f"sweepArray needs to be Python list, numpy.array or a range object. You supplied data of type: {type(inputDict['sweepArray']).__name__}"
    if type(inputDict['sweepArray']) == range:
        inputDict['sweepArray'] = [*inputDict['sweepArray']] # Unpack range into a list

    # 8. readers
    if not isinstance(inputDict['readers'],dict):
        raise ValueError(
            f'Your input dictionary with key "readers" mus be a dictionary itself. You assigned: {inputDict["readers"]} of type: {type(inputDict["readers"]).__name__}\n Please assign the instrument as "key" and the corresponding variable name(s) as values')

    # 10. Make sure

    return inputDict


# Checks if val is iterable, but not a string
def isIterable(val):
    return isinstance(val, Iterable) and not isinstance(val, str)

# Flattens a list: [[1, 2, 3], 4, [5], [6, 7]] => [1, 2, 3, 4, 5, 6, 7]
def flatten(iterable):
    flattenedList = []
    for e1 in iterable:
        if isIterable(e1):
            for e2 in e1:
                flattenedList.append(e2)
        else:
            flattenedList.append(e1)
    return flattenedList

def checkPointMatchesHeaders(point, headers): # I added heaps of flatten() in here in order to prevent weird errors where there shouldnt be any.
    if len(flatten(point)) != len(flatten(headers)):
        raise ValueError("Point {} does not have same length as header {}".format(flatten(point), flatten(headers)))

    for value, header in zip(flatten(point), flatten(headers)):
        if isIterable(header) and isIterable(value):
            if len(flatten(header)) != len(flatten(value)):
                raise ValueError("Point {} does not have same length as header {}".format(flatten(point), flatten(headers)))
        elif not isIterable(header) and not isIterable(value):
            pass
        else:
            raise ValueError("Point {} does not have same length as header {}".format(flatten(point), flatten(headers)))

def checkPlotHeaders(inputHeaders,outputHeaders,plotVars):
    if plotVars == None:
        return
    if (len(plotVars) ==2 and plotVars[0] in (flatten(inputHeaders)+flatten(outputHeaders)) and plotVars[1] in (flatten(inputHeaders)+flatten(outputHeaders))):
        return 1
    elif (len(plotVars) ==4 and plotVars[0] in (flatten(inputHeaders)+flatten(outputHeaders)) and plotVars[1] in (flatten(inputHeaders)+flatten(outputHeaders)) and plotVars[2] in (flatten(inputHeaders)+flatten(outputHeaders)) and plotVars[3] in (flatten(inputHeaders)+flatten(outputHeaders))):
        return 2
    else:
        raise ValueError("{} does either not have the right format (either two or 4 parameters) or one of the given values is not found in input or output Headers".format(plotVars))


def progressBar(current, total, barLength = 20):
    percent = float(current) * 100 / total
    arrow   = '-' * int(percent/100 * barLength - 1) + '>'
    spaces  = ' ' * (barLength - len(arrow))

    print('Progress: [%s%s] %d %%' % (arrow, spaces, percent), end='\r')


# -*- coding: utf-8 -*-
"""
VERSION 4.0
@author: Jakob Seidl
jakob.seidl@nanoelectronics.physics.unsw.edu.au
"""
import time
from numpy import random
import numpy as np
import pyneMeas.Instruments.Instrument as Instrument

@Instrument.enableOptions
class LinearNoiseGenerator(Instrument.Instrument):
    """Creates a TimeMeas virtual instrument that can set time (setter) or measure the elapsed time during a measurement.

    :param
    :return: NoiseGenerator instance

    >>> LinearNoiseGenerator()

    """


    defaultOutput = "sourceLevel"
    defaultInput = "noise"
    def __init__(self,slope = 1,std=5):
        super(LinearNoiseGenerator,self).__init__()
        self.counter  = 0
        self.type = 'NoiseGenerator'
        self.name = 'myNoiseGenerator'
        self.slope = slope
        self.std = std
    @Instrument.addOptionSetter("sourceLevel")
    def _dummySourceFunction(self,dummyVariable):
        pass

    @Instrument.addOptionGetter("noise")
    def _getNoise(self):
        self.counter += 1

        return (self.slope*self.counter + random.normal(0 , self.std))

    
    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName

    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name

    def reset(self):
        self.counter = 0

    
    def goTo(target = 0.1,delay = 0.2,stepsize = 0.2): #Dummy function. Is required for every class to be passed to the sweep functin.
        return




@Instrument.enableOptions
class SineNoiseGenerator(Instrument.Instrument):
    """Creates a TimeMeas virtual instrument that can set time (setter) or measure the elapsed time during a measurement.

    :param
    :return: NoiseGenerator instance

    >>> SineNoiseGenerator()

    """

    defaultOutput = "sourceLevel"
    defaultInput = "noise"

    def __init__(self, amplitude=2, std=0.3,linearSlope = 0):
        super(SineNoiseGenerator, self).__init__()
        self.counter = 0
        self.type = 'NoiseGenerator'
        self.name = 'myNoiseGenerator'
        self.amplitude = amplitude
        self.std = std
        self.linearSlope = linearSlope
    @Instrument.addOptionSetter("sourceLevel")
    def _dummySourceFunction(self, dummyVariable):
        pass

    @Instrument.addOptionGetter("noise")
    def _getNoise(self):
        self.counter += 1

        return (self.amplitude * np.sin((self.counter)/30*2*np.pi) + self.linearSlope*self.counter + random.normal(0, self.std))


    @Instrument.addOptionSetter("name")
    def _setName(self, instrumentName):
        self.name = instrumentName

    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name

    def goTo(target=0.1, delay=0.2,
             stepsize=0.2):  # Dummy function. Is required for every class to be passed to the sweep function.
        return
    def reset(self):
        self.counter = 0



@Instrument.enableOptions
class DoubleNoiseGenerator(Instrument.Instrument):
    """Creates a TimeMeas virtual instrument that can set time (setter) or measure the elapsed time during a measurement.

    :param
    :return: NoiseGenerator instance

    >>> DoubleNoiseGenerator()

    """

    defaultOutput = "sourceLevel"
    defaultInput = "noise"

    def __init__(self, amplitude=1, std=1,linearSlope = 0):
        super(DoubleNoiseGenerator, self).__init__()
        self.counter = 0
        self.name = 'myDoubleNoiseGenerator'
        self.amplitude = amplitude
        self.std = std
        self.type = 'DoubleNoiseGenerator'
        self.linearSlope = linearSlope
    @Instrument.addOptionSetter("sourceLevel")
    def _dummySourceFunction(self, dummyVariable):
        pass

    @Instrument.addOptionGetter("noise")
    def _getNoise(self):
        self.counter += 1
        res = self.amplitude * np.sin((self.counter)/30*2*np.pi) + self.linearSlope*self.counter + random.normal(0, self.std)
        return (res,-res)


    @Instrument.addOptionSetter("name")
    def _setName(self, instrumentName):
        self.name = instrumentName

    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name

    def reset(self):
        self.counter = 0

    def goTo(target=0.1, delay=0.2,
             stepsize=0.2):  # Dummy function. Is required for every class to be passed to the sweep functin.
        return


"""
author: Adam Micolich
Updated by Jakob Seidl

This module does the output handling for the USB-6216, which is effectively a pair of analog outputs
and a set of 8 analog inputs. The input handling is done by a separate .py. There is a settable option for 
where the voltage feedback comes from. The default feedback is internal, i.e., USB-6216 reads back what it
is putting out via an internal connection. The alternative is to T- off to a spare input (e.g., ai7) and 
direct read. 
"""

import pyneMeas.Instruments.Instrument as Instrument
import nidaqmx as nmx
import time
import numpy as np

@Instrument.enableOptions
class USB6216Out(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    defaultInput = "outputLevel"
    defaultOutput = "outputLevel"
    defaultFeedBack = "Int"

    def __init__(self, address,usbPort = 'Dev1'):
        super(USB6216Out, self).__init__()
        self.dev = address
        self.type ="USB6216"  #We can check each instrument for its type and react accordingly
        self.name = "myUSB6216"
        self.usbPort = usbPort
        if self.dev in [0, 1]:
            self.port = f"{self.usbPort}/ao{self.dev}"
            
        else:
            raise ValueError(
                f'Please insert a valid Input port for the NIDaQ ranging from 0 to 1. You entered {self.dev}')

    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
         
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name

    @Instrument.addOptionSetter("outputLevel")
    def _setOutputLevel(self,outputLevel):
        if (10.0 >= outputLevel and outputLevel >= -10.0):
            self.output = outputLevel
            with nmx.Task() as task:
                task.ao_channels.add_ao_voltage_chan(self.port)
                task.write(outputLevel)
        else:
            raise ValueError("Output outside 10V range available".format(outputLevel))
                
    @Instrument.addOptionGetter("outputLevel")
    def _getOutputLevel(self):
        with nmx.Task() as task:
            task.ai_channels.add_ai_voltage_chan(self.fbp)
            outputLevel = task.read()
        return outputLevel
        
    @Instrument.addOptionGetter("feedBack")
    def _getScaleFactor(self):
        return self.feedBack

    @Instrument.addOptionSetter("feedBack")
    def _setFeedBack(self,feedBack):
        self.feedBack = feedBack
        
    @Instrument.addOptionGetter("extPort")
    def _getExtPort(self):
        return self.fbp
    
    @Instrument.addOptionSetter("extPort")
    def _setExtPort(self,extPort):
        self.extPort = extPort
        if self.feedBack == "Int":
            if self.dev == 0:
                self.fbp = f"{self.usbPort}/_ao0_vs_aognd"
            elif self.dev == 1:
                self.fbp = f"{self.usbPort}/_ao1_vs_aognd"
        elif self.feedBack == "Ext":
            if self.extPort == 0:
                self.fbp = f"{self.usbPort}/ai0"
            elif self.extPort == 1:
                self.fbp = f"{self.usbPort}/ai1"
            elif self.extPort == 2:
                self.fbp = f"{self.usbPort}/ai2"
            elif self.extPort == 3:
                self.fbp = f"{self.usbPort}/ai3"
            elif self.extPort == 4:
                self.fbp = f"{self.usbPort}/ai4"
            elif self.extPort == 5:
                self.fbp = f"{self.usbPort}/ai5"
            elif self.extPort == 6:
                self.fbp = f"{self.usbPort}/ai6"
            elif self.extPort == 7:
                self.fbp = f"{self.usbPort}/ai7"

    @Instrument.addOptionGetter("scaleFactor")
    def _getScaleFactor(self):
        return self.scaleFactor
    
    @Instrument.addOptionSetter("scaleFactor")
    def _setScaleFactor(self,scaleFactor):
        self.scaleFactor = scaleFactor
            
    def goTo(self,target,stepsize = 0.01,delay = 0.001): # Modified from usual as APM likes linspace better
        currentOutput = self.get("outputLevel")
        count = int(abs(target-currentOutput)/stepsize) + 1
        sweepArray = np.linspace(currentOutput,target,count,endpoint=True)
        if count < 3: #Option to avoid pointless sweeps, if you're close enough, just go direct
            with nmx.Task() as task:
                task.ao_channels.add_ao_voltage_chan(self.port)
                task.write(target)
        else:
            for point in sweepArray:
                with nmx.Task() as task:
                    task.ao_channels.add_ao_voltage_chan(self.port)
                    task.write(point)
                time.sleep(delay)
            with nmx.Task() as task:
                task.ao_channels.add_ao_voltage_chan(self.port)
                task.write(target)
            
    def close(self):
        pass


import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()



setuptools.setup(
    name="pyneMeas", # Replace with your own username
    version="0.0.3",
    author="Jakob Seidl",
    author_email="jakob.seidl91@gmail.com",
    description="A simple Python 3 interface for controlling electronic measurement instruments and scripting measurement sweeps.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JakobSeidl/pyneMeas",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,  #this should tell the package to include the non-python files defined in MANIFEST.in
    python_requires='>=3.6',
    install_requires = ['numpy',#  'cycler','kiwisolver'
                        'matplotlib','nidaqmx','pandas',#'Pillow'
                        'scipy','pyvisa'] #'pytz','pyparsing','python-dateutil','six'
    
)






# -*- coding: utf-8 -*-
"""
Created on Fri Jan 5 13:47:12 2018

@author: z5168331
"""
import pyvisa as visa
import pyneMeas.Instruments.Instrument as Instrument
import numpy as np
import time

@Instrument.enableOptions
class YokogawaGS200(Instrument.Instrument):
    # Default options to set/get when the instrument is passed into the sweeper
    defaultOutput = "sourceLevel"
    defaultInput = "senseLevel"

    def __init__(self, address):
        super(YokogawaGS200, self).__init__()
        self.dev = visa.ResourceManager().open_resource("GPIB0::"+str(address)+"::INSTR")
        print(self.dev.query("*IDN?")) # Probably should query and check we have the right device        
        #self.dev.write("*RST")
        # self.dev.write("*CLS")
        self.type ="YokogawaGS200"  #We can check each instrument for its type and react accordingly
        self.name = 'myYokogawaGS200'

    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName
    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name
    
    @Instrument.addOptionSetter("outputEnable")
    def _setOutputEnable(self, enable):
        self.dev.write(":OUTP " + ("ON" if enable else "OFF"))
        
    @Instrument.addOptionSetter("sourceMode")
    def _setSourceMode(self, mode):
        if mode == "voltage":
            # Source a voltage, sense a current
            self.dev.write("SOUR:FUNC VOLT")            
        elif mode == "current":
            # Source a current, sense a voltage
            self.dev.write("SOUR:FUNC CURR")
        else:
            raise ValueError(
                "\"{}\" is not a valid source mode for the YokogawaGS200.".format(mode) +
                " The mode must either be \"voltage\" or \"current\""
            )

    @Instrument.addOptionGetter("sourceMode")
    def _getSourceMode(self):
        mode = self.dev.query("SOURCE:FUNC?")
        if mode == u"VOLT\n":
            return "voltage"
        elif mode == u"CURR\n":
            return "current"
        else:
            raise RuntimeError("Unknown source mode {}".format(mode))


    @Instrument.addOptionSetter("sourceRange")
    def _setSourceRange(self,sourceRange):  
        mode = self._getSourceMode()
        if (mode == 'voltage'):
            if float(sourceRange) in (0.01,0.1,1,10,30):
                self.dev.write("SOUR:RANG "+str(sourceRange))  
            else:
                print("Error! Accepted ranges for voltage are: 0.01, 0.1, 1, 10, 30 Volts")
        elif(mode == 'current'):    
            if float(sourceRange) in (0.001,0.01,0.1,0.2):
                self.dev.write("SOUR:RANG "+str(sourceRange))                     
            else:
                print("Error! Accepted ranges for currents are: 0.001, 0.01, 0.1, 0.2 Amperes") 


    @Instrument.addOptionGetter("sourceRange")
    def _getSourceRange(self):
        return float(self.dev.query("SOUR:RANG?"))


    @Instrument.addOptionSetter('sourceLevel')
    def _setSourceLevel(self,sourceLevel):
        Limits = 1.2*self._getSourceRange()  #the yokogawa can put out a voltage?current 20% higher thant the actual range, e.g. 12mV at the 10mV range.
        if (Limits > sourceLevel and sourceLevel > -Limits):
                self.dev.write("SOUR:LEVEL "+str(sourceLevel))
        else:
                raise ValueError(
                    "Specified source level output of \"{} V/A \" not possible in the given voltage range.".format(sourceLevel))
                
            
        
    @Instrument.addOptionGetter('sourceLevel')
    def _getSourceLevel(self):
        return float(self.dev.query("SOUR:LEVEL?"))
    
    @Instrument.addOptionSetter('compliance')
    def _setCompliance(self,compliance):
        mode = self._getSourceMode()  
        sourceRange = self._getSourceRange()
        if (mode == 'voltage' and sourceRange in (30,10,1)): # when sourcing a voltage we want current compliance
            if 0.001 <= float(compliance) <=0.1:
                self.dev.write("SOUR:PROT:CURR "+str(compliance))
            else:
                raise ValueError(
                    "Specified source protection of \"{} A \" not possible in the given voltage range. Also take into account that the YokogawaGS200 only accepts current protection for the 30,10 and 1 Volt voltage ranges.".format(compliance))
        elif (mode == 'current'): # when sourcing a current we want voltage compliance
             if 1 <= float(compliance) <=30:
                self.dev.write("SOUR:PROT:VOLT "+str(compliance))


    def goTo(self,target,stepsize= 0.002,delay=0.10):
        currentOutput = self.get('sourceLevel')
        sign = 1 if (target>currentOutput) else -1
        sweepArray = np.arange(currentOutput,target+sign*stepsize,sign*stepsize)
        for point in sweepArray:
            self.set('sourceLevel',point)
            time.sleep(delay)
        self.set('sourceLevel',target)
    
    def close(self):
            self.dev.close()


# -*- coding: utf-8 -*-
"""
VERSION 4.0
@author: Jakob Seidl
jakob.seidl@nanoelectronics.physics.unsw.edu.au
"""
import time

import pyneMeas.Instruments.Instrument as Instrument

@Instrument.enableOptions
class TimeMeas(Instrument.Instrument):
    """Creates a TimeMeas virtual instrument that can set time (setter) or measure the elapsed time during a measurement.

    :param
    :return: TimeMeas instance

    >>> TimeMeas(0.5)
    <Instruments.TimeMeas.TimeMeas object at 0x10fcc5790>
    """


    defaultOutput = "sourceLevel"
    defaultInput = "timeStamp"
    def __init__(self):
        super(TimeMeas,self).__init__()
        # self.timeInterval = timeInterval
        self.initTime = time.time()
        self.type = 'TimeMeas'
        self.name = 'myTimeMeas'
        self.dummyPoint = 0
    
    @Instrument.addOptionSetter("sourceLevel")
    def _dummySourceFunction(self,dummyVariable):
        # time.sleep(self.timeInterval)
        self.dummyPoint = dummyVariable

    @Instrument.addOptionGetter("sourceLevel")
    def _dummyReadSourceFunction(self):
        # time.sleep(self.timeInterval)
        return self.dummyPoint

    @Instrument.addOptionGetter("type")
    def _getInstrumentType(self):
        return self.type
    @Instrument.addOptionGetter("timeStamp")
    def _getTime(self):
        return time.time()-self.initTime
    
    @Instrument.addOptionSetter("name")
    def _setName(self,instrumentName):
         self.name = instrumentName

    @Instrument.addOptionGetter("name")
    def _getName(self):
        return self.name

    @Instrument.addOptionSetter("timeStampReset")
    def _setTimeReset(self):
        """ By default, TimeMeas measures the time passed since it's creation in the beginning of a script. For more accurate time values,
            execute myTimeMeas.set("timeStampReset") before executing the sweepAndSave(...) function"""
        self.initTime = time.time()
    
    def goTo(target = 0.1,delay = 0.2,stepsize = 0.2): #Dummy function. Is required for every class to be passed to the sweep functin.
        return




"""
This module reads, supplies and updates the current 'measurement ID', a running number on each setup.
A two letter identifier of each setup, e.g., AT for Attocube, Pr for probe station, He for Heliox  etc. is also saved in this file.
Both are read in the sweep function sweepAndSave() each time a measurement is done.
"""
# This has to be set every time pyNe is installed on a new system. It denotes the path where the unique number is saved.
import platform
import os
import json
# preFix = {'Darwin':'', # THis just gives the right prefix for MAc ('Darwin' and Windows)
#           'Windows':'../'}

relPath = os.path.realpath(__file__)[:-15] #Giving the full path without the GlobalMeasID.py script ending
filePath =  relPath + 'GlobalMeasIDBinary'




def init(preFix = 'A',ID= 0):
    with open(filePath, 'w') as file:
     file.write(json.dumps({'currentPreFix':preFix,preFix:ID
                           }))

def  addPrefix(newPreFix):
    with open(filePath, 'r') as file:
        # print(file.read())
        inputDic=  json.loads(file.read())
        inputDic = {newPreFix:0,**inputDic}
    with open(filePath, 'w') as file:
        file.write(json.dumps(inputDic))
        print(f'Succesfully added the new measurement preFix/Setup: {newPreFix}')


def readCurrentID():
    with open(filePath, 'r') as file:
        Dict = json.loads(file.read())
        return Dict[Dict['currentPreFix']]

def listIDs():
    with open(filePath, 'r') as file:
        Dict = json.loads(file.read())
        current = f"Currently used Prefix/Setup: {Dict['currentPreFix']}  --> ID = {Dict[Dict['currentPreFix']]} \n -------------------------- \nOther available Setups/Prefixes are: "
        print(current)
        retString = ''
        retString = retString.join(current+'\n')
        for key,item in Dict.items():
            if (key !='currentPreFix' and key != Dict['currentPreFix']):
                print(f"Prefix/Setup: {key}  --> ID = {item}")
                retString = retString.join(f"Prefix/Setup: {key}  --> ID = {item}\n")

        # return retString
def increaseID():
    with open(filePath, 'r') as file:
        # print(file.read())
        inputDic=  json.loads(file.read())
        inputDic[inputDic['currentPreFix']] = str(int(inputDic[inputDic['currentPreFix']]) + 1)
    with open(filePath, 'w') as file:
        file.write(json.dumps(inputDic))

def setCurrentSetup(preFix):
    with open(filePath, 'r') as file:
        Dict = json.loads(file.read())
        previousPrefix = Dict['currentPreFix']
    if preFix in Dict.keys():

        if previousPrefix == preFix:
            print(f'Using current Id/Prefix: {previousPrefix}')
            pass
        else:
            Dict['currentPreFix'] = preFix
            with open(filePath, 'w') as file:
                file.write(json.dumps(Dict))
                print(f'Succesfully changed preFix/Setup from: {previousPrefix} ---> {preFix}')
    else:
        print(listIDs())
        raise Exception(f'Prefix not defined!!\n Currently defined prefixes can be listed by usign the listIDs() function.\n Define the desired prefix first using the addPrefix(newPrefix) method')

def readCurrentSetup():
    with open(filePath, 'r') as file:
        Dict=  json.loads(file.read())
        return Dict['currentPreFix']


#def readCurrentID():
#        IDTXT =open(IDpath,"r")
#        ID = int(IDTXT.read())
#        IDTXT.close()
#        return(ID) 
#    
#def increaseID():
#        ID = readCurrentID()
#        IDTXT =open(IDpath,"w")
#        IDTXT.write(str(ID+1))
#        IDTXT.close()
#def readCurrentSetup():
#    return currentSetup
#        
#def init():
#    """Creates a new ID file at the path specified in variable 'Idpath'. Should only be called if the ID file was deleted etc. """
#    IDTXT =open(IDpath,"w")
#    IDTXT.write(str(0))
#    IDTXT.close()
    